__version__ = '0.1.40'
__title__ = "knowsmore"
__description__ = "KnowsMore is a swiss army knife tool for pentesting Microsoft Active Directory (NTLM Hashes, BloodHound, NTDS and DCSync)."
__url__ = "https://github.com/helviojunior/knowsmore"
__build__ = 0x63c974a8
__author__ = "Helvio Junior  (M4v3r1ck)"
__author_email__ = "helvio.junior@sec4us.com.br"
__license__ = "GPL-3.0"
__copyright__ = "Copyright Helvio Junior"
